@@ -1,46 +1,25 @@
 /*
  * security/ccsecurity/domain.c
  *
- * Copyright (C) 2005-2010  NTT DATA CORPORATION
- *
- * Version: 1.8.0-pre   2010/10/25
- *
- * This file is applicable to both 2.4.30 and 2.6.11 and later.
- * See README.ccs for ChangeLog.
+ * Copyright (C) 2005-2011  NTT DATA CORPORATION
  *
+ * Version: 1.8.2   2011/06/20
  */
 
 #include "internal.h"
 
-/*
- * The global domains referred by "use_group" keyword.
- *
- * Although "use_group" needs only "struct list_head acl_info_list[2]",
- * we define structure for "use_group" as "struct ccs_domain_info" in order to
- * use common code.
- */
-struct ccs_domain_info ccs_acl_group[CCS_MAX_ACL_GROUPS];
-
 /* The initial domain. */
 struct ccs_domain_info ccs_kernel_domain;
 
 /* The list for "struct ccs_domain_info". */
 LIST_HEAD(ccs_domain_list);
 
-/* List of domain policy. */
-struct list_head ccs_policy_list[CCS_MAX_POLICY];
-/* List of "struct ccs_group". */
-struct list_head ccs_group_list[CCS_MAX_GROUP];
-/* List of "struct ccs_condition" and "struct ccs_ipv6addr". */
-struct list_head ccs_shared_list[CCS_MAX_LIST];
-
 /**
  * ccs_update_policy - Update an entry for exception policy.
  *
  * @new_entry:       Pointer to "struct ccs_acl_info".
  * @size:            Size of @new_entry in bytes.
- * @is_delete:       True if it is a delete request.
- * @list:            Pointer to "struct list_head".
+ * @param:           Pointer to "struct ccs_acl_param".
  * @check_duplicate: Callback function to find duplicated entry.
  *
  * Returns 0 on success, negative value otherwise.
@@ -48,22 +27,23 @@ struct list_head ccs_shared_list[CCS_MAX_LIST];
  * Caller holds ccs_read_lock().
  */
 int ccs_update_policy(struct ccs_acl_head *new_entry, const int size,
-		      bool is_delete, struct list_head *list,
+		      struct ccs_acl_param *param,
 		      bool (*check_duplicate) (const struct ccs_acl_head *,
 					       const struct ccs_acl_head *))
 {
-	int error = is_delete ? -ENOENT : -ENOMEM;
+	int error = param->is_delete ? -ENOENT : -ENOMEM;
 	struct ccs_acl_head *entry;
+	struct list_head *list = param->list;
 	if (mutex_lock_interruptible(&ccs_policy_lock))
 		return -ENOMEM;
 	list_for_each_entry_srcu(entry, list, list, &ccs_ss) {
 		if (!check_duplicate(entry, new_entry))
 			continue;
-		entry->is_deleted = is_delete;
+		entry->is_deleted = param->is_delete;
 		error = 0;
 		break;
 	}
-	if (error && !is_delete) {
+	if (error && !param->is_delete) {
 		entry = ccs_commit_ok(new_entry, size);
 		if (entry) {
 			list_add_tail_rcu(&entry->list, list);
@@ -109,23 +89,18 @@ int ccs_update_domain(struct ccs_acl_info *new_entry, const int size,
 					       struct ccs_acl_info *,
 					       const bool))
 {
-	int error = param->is_delete ? -ENOENT : -ENOMEM;
-	struct ccs_acl_info *entry;
-	const u8 type = new_entry->type;
-	const u8 i = type == CCS_TYPE_AUTO_EXECUTE_HANDLER ||
-		type == CCS_TYPE_DENIED_EXECUTE_HANDLER ||
-		type == CCS_TYPE_AUTO_TASK_ACL;
 	const bool is_delete = param->is_delete;
-	struct ccs_domain_info * const domain = param->domain;
+	int error = is_delete ? -ENOENT : -ENOMEM;
+	struct ccs_acl_info *entry;
+	struct list_head * const list = param->list;
 	if (param->data[0]) {
-		new_entry->cond = ccs_get_condition(param->data);
+		new_entry->cond = ccs_get_condition(param);
 		if (!new_entry->cond)
 			return -EINVAL;
 	}
 	if (mutex_lock_interruptible(&ccs_policy_lock))
 		goto out;
-	list_for_each_entry_srcu(entry, &domain->acl_info_list[i], list,
-				 &ccs_ss) {
+	list_for_each_entry_srcu(entry, list, list, &ccs_ss) {
 		if (!ccs_same_acl_head(entry, new_entry) ||
 		    !check_duplicate(entry, new_entry))
 			continue;
@@ -140,8 +115,7 @@ int ccs_update_domain(struct ccs_acl_info *new_entry, const int size,
 	if (error && !is_delete) {
 		entry = ccs_commit_ok(new_entry, size);
 		if (entry) {
-			list_add_tail_rcu(&entry->list,
-					  &domain->acl_info_list[i]);
+			list_add_tail_rcu(&entry->list, list);
 			error = 0;
 		}
 	}
@@ -169,10 +143,9 @@ void ccs_check_acl(struct ccs_request_info *r,
 	const struct ccs_domain_info *domain = ccs_current_domain();
 	struct ccs_acl_info *ptr;
 	bool retried = false;
-	const u8 i = !check_entry;
+	const struct list_head *list = &domain->acl_info_list;
 retry:
-	list_for_each_entry_srcu(ptr, &domain->acl_info_list[i], list,
-				 &ccs_ss) {
+	list_for_each_entry_srcu(ptr, list, list, &ccs_ss) {
 		if (ptr->is_deleted)
 			continue;
 		if (ptr->type != r->param_type)
@@ -187,7 +160,7 @@ retry:
 	}
 	if (!retried) {
 		retried = true;
-		domain = &ccs_acl_group[domain->group];
+		list = &domain->ns->acl_group[domain->group];
 		goto retry;
 	}
 	r->granted = false;
@@ -214,22 +187,27 @@ static bool ccs_same_transition_control(const struct ccs_acl_head *a,
 }
 
 /**
- * ccs_update_transition_control_entry - Update "struct ccs_transition_control" list.
+ * ccs_write_transition_control - Write "struct ccs_transition_control" list.
  *
- * @domainname: The name of domain. Maybe NULL.
- * @program:    The name of program. Maybe NULL.
- * @type:       Type of transition.
- * @is_delete:  True if it is a delete request.
+ * @param: Pointer to "struct ccs_acl_param".
+ * @type:  Type of this entry.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_update_transition_control_entry(const char *domainname,
-					       const char *program,
-					       const u8 type,
-					       const bool is_delete)
+int ccs_write_transition_control(struct ccs_acl_param *param, const u8 type)
 {
 	struct ccs_transition_control e = { .type = type };
-	int error = is_delete ? -ENOENT : -ENOMEM;
+	int error = param->is_delete ? -ENOENT : -ENOMEM;
+	char *program = param->data;
+	char *domainname = strstr(program, " from ");
+	if (domainname) {
+		*domainname = '\0';
+		domainname += 6;
+	} else if (type == CCS_TRANSITION_CONTROL_NO_KEEP ||
+		   type == CCS_TRANSITION_CONTROL_KEEP) {
+		domainname = program;
+		program = NULL;
+	}
 	if (program && strcmp(program, "any")) {
 		if (!ccs_correct_path(program))
 			return -EINVAL;
@@ -247,8 +225,8 @@ static int ccs_update_transition_control_entry(const char *domainname,
 		if (!e.domainname)
 			goto out;
 	}
-	error = ccs_update_policy(&e.head, sizeof(e), is_delete,
-				  &ccs_policy_list[CCS_ID_TRANSITION_CONTROL],
+	param->list = &param->ns->policy_list[CCS_ID_TRANSITION_CONTROL];
+	error = ccs_update_policy(&e.head, sizeof(e), param,
 				  ccs_same_transition_control);
 out:
 	ccs_put_name(e.domainname);
@@ -256,31 +234,6 @@ out:
 	return error;
 }
 
-/**
- * ccs_write_transition_control - Write "struct ccs_transition_control" list.
- *
- * @data:      String to parse.
- * @is_delete: True if it is a delete request.
- * @type:      Type of this entry.
- *
- * Returns 0 on success, negative value otherwise.
- */
-int ccs_write_transition_control(char *data, const bool is_delete,
-				 const u8 type)
-{
-	char *domainname = strstr(data, " from ");
-	if (domainname) {
-		*domainname = '\0';
-		domainname += 6;
-	} else if (type == CCS_TRANSITION_CONTROL_NO_KEEP ||
-		   type == CCS_TRANSITION_CONTROL_KEEP) {
-		domainname = data;
-		data = NULL;
-	}
-	return ccs_update_transition_control_entry(domainname, data, type,
-						   is_delete);
-}
-
 /**
  * ccs_last_word - Get last component of a domainname.
  *
@@ -297,58 +250,88 @@ static const char *ccs_last_word(const char *name)
 }
 
 /**
- * ccs_transition_type - Get domain transition type.
+ * ccs_scan_transition - Try to find specific domain transition type.
  *
- * @domainname: The name of domain.
- * @program:    The name of program.
+ * @list:       Pointer to "struct list_head".
+ * @domainname: The name of current domain.
+ * @program:    The name of requested program.
+ * @last_name:  The last component of @domainname.
+ * @type:       One of values in "enum ccs_transition_type".
  *
- * Returns CCS_TRANSITION_CONTROL_INITIALIZE if executing @program
- * reinitializes domain transition, CCS_TRANSITION_CONTROL_KEEP if executing
- * @program suppresses domain transition, others otherwise.
+ * Returns true if found one, false otherwise.
  *
  * Caller holds ccs_read_lock().
  */
-static u8 ccs_transition_type(const struct ccs_path_info *domainname,
-			      const struct ccs_path_info *program)
+static inline bool ccs_scan_transition(const struct list_head *list,
+				       const struct ccs_path_info *domainname,
+				       const struct ccs_path_info *program,
+				       const char *last_name,
+				       const enum ccs_transition_type type)
 {
 	const struct ccs_transition_control *ptr;
-	const char *last_name = ccs_last_word(domainname->name);
-	u8 type;
-	for (type = 0; type < CCS_MAX_TRANSITION_TYPE; type++) {
-next:
-		list_for_each_entry_srcu(ptr, &ccs_policy_list
-					 [CCS_ID_TRANSITION_CONTROL],
-					 head.list, &ccs_ss) {
-			if (ptr->head.is_deleted || ptr->type != type)
-				continue;
-			if (ptr->domainname) {
-				if (!ptr->is_last_name) {
-					if (ptr->domainname != domainname)
-						continue;
-				} else {
-					/*
-					 * Use direct strcmp() since this is
-					 * unlikely used.
-					 */
-					if (strcmp(ptr->domainname->name,
-						   last_name))
-						continue;
-				}
-			}
-			if (ptr->program && ccs_pathcmp(ptr->program, program))
-				continue;
-			if (type == CCS_TRANSITION_CONTROL_NO_INITIALIZE) {
+	list_for_each_entry_srcu(ptr, list, head.list, &ccs_ss) {
+		if (ptr->head.is_deleted || ptr->type != type)
+			continue;
+		if (ptr->domainname) {
+			if (!ptr->is_last_name) {
+				if (ptr->domainname != domainname)
+					continue;
+			} else {
 				/*
-				 * Do not check for initialize_domain if
-				 * no_initialize_domain matched.
+				 * Use direct strcmp() since this is
+				 * unlikely used.
 				 */
-				type = CCS_TRANSITION_CONTROL_NO_KEEP;
-				goto next;
+				if (strcmp(ptr->domainname->name, last_name))
+					continue;
 			}
-			goto done;
 		}
+		if (ptr->program && ccs_pathcmp(ptr->program, program))
+			continue;
+		return true;
+	}
+	return false;
+}
+
+/**
+ * ccs_transition_type - Get domain transition type.
+ *
+ * @ns:         Pointer to "struct ccs_policy_namespace".
+ * @domainname: The name of current domain.
+ * @program:    The name of requested program.
+ *
+ * Returns CCS_TRANSITION_CONTROL_TRANSIT if executing @program causes domain
+ * transition across namespaces, CCS_TRANSITION_CONTROL_INITIALIZE if executing
+ * @program reinitializes domain transition within that namespace,
+ * CCS_TRANSITION_CONTROL_KEEP if executing @program stays at @domainname ,
+ * others otherwise.
+ *
+ * Caller holds ccs_read_lock().
+ */
+static enum ccs_transition_type ccs_transition_type
+(const struct ccs_policy_namespace *ns, const struct ccs_path_info *domainname,
+ const struct ccs_path_info *program)
+{
+	const char *last_name = ccs_last_word(domainname->name);
+	enum ccs_transition_type type = CCS_TRANSITION_CONTROL_NO_RESET;
+	while (type < CCS_MAX_TRANSITION_TYPE) {
+		const struct list_head * const list =
+			&ns->policy_list[CCS_ID_TRANSITION_CONTROL];
+		if (!ccs_scan_transition(list, domainname, program, last_name,
+					 type)) {
+			type++;
+			continue;
+		}
+		if (type != CCS_TRANSITION_CONTROL_NO_RESET &&
+		    type != CCS_TRANSITION_CONTROL_NO_INITIALIZE)
+			break;
+		/*
+		 * Do not check for reset_domain if no_reset_domain matched.
+		 * Do not check for initialize_domain if no_initialize_domain
+		 * matched.
+		 */
+		type++;
+		type++;
 	}
-done:
 	return type;
 }
 
@@ -370,20 +353,18 @@ static bool ccs_same_aggregator(const struct ccs_acl_head *a,
 }
 
 /**
- * ccs_update_aggregator_entry - Update "struct ccs_aggregator" list.
+ * ccs_write_aggregator - Write "struct ccs_aggregator" list.
  *
- * @original_name:   The original program's name.
- * @aggregated_name: The aggregated program's name.
- * @is_delete:       True if it is a delete request.
+ * @param: Pointer to "struct ccs_acl_param".
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_update_aggregator_entry(const char *original_name,
-				       const char *aggregated_name,
-				       const bool is_delete)
+int ccs_write_aggregator(struct ccs_acl_param *param)
 {
 	struct ccs_aggregator e = { };
-	int error = is_delete ? -ENOENT : -ENOMEM;
+	int error = param->is_delete ? -ENOENT : -ENOMEM;
+	const char *original_name = ccs_read_token(param);
+	const char *aggregated_name = ccs_read_token(param);
 	if (!ccs_correct_word(original_name) ||
 	    !ccs_correct_path(aggregated_name))
 		return -EINVAL;
@@ -392,8 +373,8 @@ static int ccs_update_aggregator_entry(const char *original_name,
 	if (!e.original_name || !e.aggregated_name ||
 	    e.aggregated_name->is_patterned) /* No patterns allowed. */
 		goto out;
-	error = ccs_update_policy(&e.head, sizeof(e), is_delete,
-				  &ccs_policy_list[CCS_ID_AGGREGATOR],
+	param->list = &param->ns->policy_list[CCS_ID_AGGREGATOR];
+	error = ccs_update_policy(&e.head, sizeof(e), param,
 				  ccs_same_aggregator);
 out:
 	ccs_put_name(e.original_name);
@@ -401,60 +382,95 @@ out:
 	return error;
 }
 
+/* Domain create handler. */
+
 /**
- * ccs_write_aggregator - Write "struct ccs_aggregator" list.
+ * ccs_find_namespace - Find specified namespace.
  *
- * @data:      String to parse.
- * @is_delete: True if it is a delete request.
+ * @name: Name of namespace to find.
+ * @len:  Length of @name.
  *
- * Returns 0 on success, negative value otherwise.
+ * Returns pointer to "struct ccs_policy_namespace" if found, NULL otherwise.
+ *
+ * Caller holds ccs_read_lock().
  */
-int ccs_write_aggregator(char *data, const bool is_delete)
+static struct ccs_policy_namespace *ccs_find_namespace(const char *name,
+						       const unsigned int len)
 {
-	char *w[2];
-	if (!ccs_tokenize(data, w, sizeof(w)) || !w[1][0])
-		return -EINVAL;
-	return ccs_update_aggregator_entry(w[0], w[1], is_delete);
+	struct ccs_policy_namespace *ns;
+	list_for_each_entry_srcu(ns, &ccs_namespace_list, namespace_list,
+				 &ccs_ss) {
+		if (strncmp(name, ns->name, len) ||
+		    (name[len] && name[len] != ' '))
+			continue;
+		return ns;
+	}
+	return NULL;
 }
 
-/* Domain create/delete handler. */
 
 /**
- * ccs_delete_domain - Delete a domain.
+ * ccs_assign_namespace - Create a new namespace.
  *
- * @domainname: The name of domain.
+ * @domainname: Name of namespace to create.
  *
- * Returns 0.
+ * Returns pointer to "struct ccs_policy_namespace" on success, NULL otherwise.
+ *
+ * Caller holds ccs_read_lock().
  */
-int ccs_delete_domain(char *domainname)
+struct ccs_policy_namespace *ccs_assign_namespace(const char *domainname)
 {
-	struct ccs_domain_info *domain;
-	struct ccs_path_info name;
-	name.name = domainname;
-	ccs_fill_path_info(&name);
+	struct ccs_policy_namespace *ptr;
+	struct ccs_policy_namespace *entry;
+	const char *cp = domainname;
+	unsigned int len = 0;
+	while (*cp && *cp++ != ' ')
+		len++;
+	ptr = ccs_find_namespace(domainname, len);
+	if (ptr)
+		return ptr;
+	if (len >= CCS_EXEC_TMPSIZE - 10 || !ccs_domain_def(domainname))
+		return NULL;
+	entry = kzalloc(sizeof(*entry) + len + 1, CCS_GFP_FLAGS);
+	if (!entry)
+		return NULL;
 	if (mutex_lock_interruptible(&ccs_policy_lock))
-		return 0;
-	/* Is there an active domain? */
-	list_for_each_entry_srcu(domain, &ccs_domain_list, list, &ccs_ss) {
-		/* Never delete ccs_kernel_domain. */
-		if (domain == &ccs_kernel_domain)
-			continue;
-		if (domain->is_deleted ||
-		    ccs_pathcmp(domain->domainname, &name))
-			continue;
-		domain->is_deleted = true;
-		break;
+		goto out;
+	ptr = ccs_find_namespace(domainname, len);
+	if (!ptr && ccs_memory_ok(entry, sizeof(*entry) + len + 1)) {
+		char *name = (char *) (entry + 1);
+		ptr = entry;
+		memmove(name, domainname, len);
+		name[len] = '\0';
+		entry->name = name;
+		ccs_init_policy_namespace(entry);
+		entry = NULL;
 	}
 	mutex_unlock(&ccs_policy_lock);
-	return 0;
+out:
+	kfree(entry);
+	return ptr;
 }
 
 /**
- * ccs_assign_domain - Create a domain.
+ * ccs_namespace_jump - Check for namespace jump.
+ *
+ * @domainname: Name of domain.
+ *
+ * Returns true if namespace differs, false otherwise.
+ */
+static bool ccs_namespace_jump(const char *domainname)
+{
+	const char *namespace = ccs_current_namespace()->name;
+	const int len = strlen(namespace);
+	return strncmp(domainname, namespace, len) ||
+		(domainname[len] && domainname[len] != ' ');
+}
+
+/**
+ * ccs_assign_domain - Create a domain or a namespace.
  *
  * @domainname: The name of domain.
- * @profile:    Profile number to assign if the domain was newly created.
- * @group:      Group number to assign if the domain was newly created.
  * @transit:    True if transit to domain found or created.
  *
  * Returns pointer to "struct ccs_domain_info" on success, NULL otherwise.
@@ -462,19 +478,53 @@ int ccs_delete_domain(char *domainname)
  * Caller holds ccs_read_lock().
  */
 struct ccs_domain_info *ccs_assign_domain(const char *domainname,
-					  const u8 profile, const u8 group,
 					  const bool transit)
 {
+	struct ccs_security *security = ccs_current_security();
 	struct ccs_domain_info e = { };
 	struct ccs_domain_info *entry = ccs_find_domain(domainname);
 	bool created = false;
-	if (entry)
-		goto out;
+	if (entry) {
+		if (transit) {
+			/*
+			 * Since namespace is created at runtime, profiles may
+			 * not be created by the moment the process transits to
+			 * that domain. Do not perform domain transition if
+			 * profile for that domain is not yet created.
+			 */
+			if (ccs_policy_loaded &&
+			    !entry->ns->profile_ptr[entry->profile])
+				return NULL;
+			security->ccs_domain_info = entry;
+		}
+		return entry;
+	}
+	/* Requested domain does not exist. */
+	/* Don't create requested domain if domainname is invalid. */
 	if (strlen(domainname) >= CCS_EXEC_TMPSIZE - 10 ||
 	    !ccs_correct_domain(domainname))
 		return NULL;
-	e.profile = profile;
-	e.group = group;
+	/*
+	 * Since definition of profiles and acl_groups may differ across
+	 * namespaces, do not inherit "use_profile" and "use_group" settings
+	 * by automatically creating requested domain upon domain transition.
+	 */
+	if (transit && ccs_namespace_jump(domainname))
+		return NULL;
+	e.ns = ccs_assign_namespace(domainname);
+	if (!e.ns)
+		return NULL;
+	/*
+	 * "use_profile" and "use_group" settings for automatically created
+	 * domains are inherited from current domain. These are 0 for manually
+	 * created domains.
+	 */
+	if (transit) {
+		const struct ccs_domain_info *domain =
+			security->ccs_domain_info;
+		e.profile = domain->profile;
+		e.group = domain->group;
+	}
 	e.domainname = ccs_get_name(domainname);
 	if (!e.domainname)
 		return NULL;
@@ -484,8 +534,7 @@ struct ccs_domain_info *ccs_assign_domain(const char *domainname,
 	if (!entry) {
 		entry = ccs_commit_ok(&e, sizeof(e));
 		if (entry) {
-			INIT_LIST_HEAD(&entry->acl_info_list[0]);
-			INIT_LIST_HEAD(&entry->acl_info_list[1]);
+			INIT_LIST_HEAD(&entry->acl_info_list);
 			list_add_tail_rcu(&entry->list, &ccs_domain_list);
 			created = true;
 		}
@@ -494,13 +543,13 @@ struct ccs_domain_info *ccs_assign_domain(const char *domainname,
 out:
 	ccs_put_name(e.domainname);
 	if (entry && transit) {
-		ccs_current_security()->ccs_domain_info = entry;
+		security->ccs_domain_info = entry;
 		if (created) {
 			struct ccs_request_info r;
 			ccs_init_request_info(&r, CCS_MAC_FILE_EXECUTE);
 			r.granted = false;
-			ccs_write_log(&r, "use_profile %u\n", profile);
-			ccs_write_log(&r, "use_group %u\n", group);
+			ccs_write_log(&r, "use_profile %u\n", entry->profile);
+			ccs_write_log(&r, "use_group %u\n", entry->group);
 			ccs_update_stat(CCS_STAT_POLICY_UPDATES);
 		}
 	}
@@ -527,6 +576,7 @@ static int ccs_find_next_domain(struct ccs_execve *ee)
 	struct ccs_path_info rn = { }; /* real name */
 	int retval;
 	bool need_kfree = false;
+	bool reject_on_transition_failure = false;
 retry:
 	r->matched_acl = NULL;
 	if (need_kfree) {
@@ -553,10 +603,10 @@ retry:
 		}
 	} else {
 		struct ccs_aggregator *ptr;
+		struct list_head *list =
+			&old_domain->ns->policy_list[CCS_ID_AGGREGATOR];
 		/* Check 'aggregator' directive. */
-		list_for_each_entry_srcu(ptr,
-					 &ccs_policy_list[CCS_ID_AGGREGATOR],
-					 head.list, &ccs_ss) {
+		list_for_each_entry_srcu(ptr, list, head.list, &ccs_ss) {
 			if (ptr->head.is_deleted ||
 			    !ccs_path_matches_pattern(&rn, ptr->original_name))
 				continue;
@@ -589,11 +639,21 @@ retry:
 	}
 
 	/* Calculate domain to transit to. */
-	switch (ccs_transition_type(old_domain->domainname, &rn)) {
+	switch (ccs_transition_type(old_domain->ns, old_domain->domainname,
+				    &rn)) {
+	case CCS_TRANSITION_CONTROL_RESET:
+		/* Transit to the root of specified namespace. */
+		snprintf(ee->tmp, CCS_EXEC_TMPSIZE - 1, "<%s>", rn.name);
+		/*
+		 * Make do_execve() fail if domain transition across namespaces
+		 * has failed.
+		 */
+		reject_on_transition_failure = true;
+		break;
 	case CCS_TRANSITION_CONTROL_INITIALIZE:
-		/* Transit to the child of ccs_kernel_domain domain. */
-		snprintf(ee->tmp, CCS_EXEC_TMPSIZE - 1, CCS_ROOT_NAME " " "%s",
-			 rn.name);
+		/* Transit to the child of current namespace's root. */
+		snprintf(ee->tmp, CCS_EXEC_TMPSIZE - 1, "%s %s",
+			 old_domain->ns->name, rn.name);
 		break;
 	case CCS_TRANSITION_CONTROL_KEEP:
 		/* Keep current domain. */
@@ -624,7 +684,7 @@ retry:
 	 * Make task->ccs_flags visible to GC before changing
 	 * task->ccs_domain_info.
 	 */
-	smp_mb();
+	smp_wmb();
 	/*
 	 * Proceed to the next domain in order to allow reaching via PID.
 	 * It will be reverted if execve() failed. Reverting is not good.
@@ -632,11 +692,14 @@ retry:
 	 * enforcing mode.
 	 */
 	if (!domain)
-		domain = ccs_assign_domain(ee->tmp, r->profile,
-					   old_domain->group, true);
+		domain = ccs_assign_domain(ee->tmp, true);
 	if (domain)
 		retval = 0;
-	else if (r->mode == CCS_CONFIG_ENFORCING)
+	else if (reject_on_transition_failure) {
+		printk(KERN_WARNING
+		       "ERROR: Domain '%s' not ready.\n", ee->tmp);
+		retval = -ENOMEM;
+	} else if (r->mode == CCS_CONFIG_ENFORCING)
 		retval = -ENOMEM;
 	else {
 		retval = 0;
@@ -666,7 +729,7 @@ static int ccs_environ(struct ccs_execve *ee)
 {
 	struct ccs_request_info *r = &ee->r;
 	struct linux_binprm *bprm = ee->bprm;
-	/* env_page->data is allocated by ccs_dump_page(). */
+	/* env_page.data is allocated by ccs_dump_page(). */
 	struct ccs_page_dump env_page = { };
 	char *arg_ptr; /* Size is CCS_EXEC_TMPSIZE bytes */
 	int arg_len = 0;
@@ -677,8 +740,8 @@ static int ccs_environ(struct ccs_execve *ee)
 	/* printk(KERN_DEBUG "start %d %d\n", argv_count, envp_count); */
 	int error = -ENOMEM;
 	ee->r.type = CCS_MAC_ENVIRON;
-	ee->r.mode = ccs_get_mode(ccs_current_domain()->profile,
-				  CCS_MAC_ENVIRON);
+	ee->r.profile = ccs_current_domain()->profile;
+	ee->r.mode = ccs_get_mode(ee->r.profile, CCS_MAC_ENVIRON);
 	if (!r->mode || !envp_count)
 		return 0;
 	arg_ptr = kzalloc(CCS_EXEC_TMPSIZE, CCS_GFP_FLAGS);
@@ -732,7 +795,7 @@ static int ccs_environ(struct ccs_execve *ee)
 		offset = 0;
 	}
 out:
-	if (r->mode != 3)
+	if (r->mode != CCS_CONFIG_ENFORCING)
 		error = 0;
 	kfree(env_page.data);
 	kfree(arg_ptr);
@@ -893,7 +956,6 @@ static int ccs_try_alt_exec(struct ccs_execve *ee)
 	int retval;
 	const int original_argc = bprm->argc;
 	const int original_envc = bprm->envc;
-	struct ccs_security *task = ccs_current_security();
 
 	/* Close the requested program's dentry. */
 	ee->obj.path1.dentry = NULL;
@@ -1012,7 +1074,7 @@ static int ccs_try_alt_exec(struct ccs_execve *ee)
 			handler_len - root_len);
 		ccs_unescape(cp);
 		retval = -ENOENT;
-		if (!*cp || *cp != '/')
+		if (*cp != '/')
 			goto out;
 		retval = ccs_copy_argv(cp, bprm);
 		if (retval < 0)
@@ -1042,9 +1104,9 @@ static int ccs_try_alt_exec(struct ccs_execve *ee)
 	retval = prepare_binprm(bprm);
 	if (retval < 0)
 		goto out;
-	task->ccs_flags |= CCS_DONT_SLEEP_ON_ENFORCE_ERROR;
+	ee->r.dont_sleep_on_enforce_error = true;
 	retval = ccs_find_next_domain(ee);
-	task->ccs_flags &= ~CCS_DONT_SLEEP_ON_ENFORCE_ERROR;
+	ee->r.dont_sleep_on_enforce_error = false;
 out:
 	return retval;
 }
@@ -1205,7 +1267,7 @@ void ccs_finish_execve(int retval, struct ccs_execve *ee)
 		 * Make task->ccs_domain_info visible to GC before changing
 		 * task->ccs_flags.
 		 */
-		smp_mb();
+		smp_wmb();
 	} else {
 		/* Mark the current process as execute handler. */
 		if (ee->handler)
@@ -1244,8 +1306,10 @@ static int __ccs_search_binary_handler(struct linux_binprm *bprm,
 {
 	struct ccs_execve *ee;
 	int retval;
+#ifndef CONFIG_CCSECURITY_OMIT_USERSPACE_LOADER
 	if (!ccs_policy_loaded)
 		ccsecurity_exports.load_policy(bprm->filename);
+#endif
 	retval = ccs_start_execve(bprm, &ee);
 	if (!retval)
 		retval = search_binary_handler(bprm, regs);
