@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2005-2011  NTT DATA CORPORATION
  *
- * Version: 1.8.1   2011/04/01
+ * Version: 1.8.2   2011/06/20
  */
 
 #include "internal.h"
@@ -79,16 +79,12 @@ const u8 ccs_pn2mac[CCS_MAX_PATH_NUMBER_OPERATION] = {
  */
 void ccs_put_name_union(struct ccs_name_union *ptr)
 {
-	if (!ptr)
-		return;
-	if (ptr->is_group)
-		ccs_put_group(ptr->group);
-	else
-		ccs_put_name(ptr->filename);
+	ccs_put_group(ptr->group);
+	ccs_put_name(ptr->filename);
 }
 
 /**
- * ccs_put_name_union - Drop reference on "struct ccs_number_union".
+ * ccs_put_number_union - Drop reference on "struct ccs_number_union".
  *
  * @ptr: Pointer to "struct ccs_number_union".
  *
@@ -96,8 +92,7 @@ void ccs_put_name_union(struct ccs_name_union *ptr)
  */
 void ccs_put_number_union(struct ccs_number_union *ptr)
 {
-	if (ptr && ptr->is_group)
-		ccs_put_group(ptr->group);
+	ccs_put_group(ptr->group);
 }
 
 /**
@@ -111,7 +106,7 @@ void ccs_put_number_union(struct ccs_number_union *ptr)
 bool ccs_compare_number_union(const unsigned long value,
 			      const struct ccs_number_union *ptr)
 {
-	if (ptr->is_group)
+	if (ptr->group)
 		return ccs_number_matches_group(value, value, ptr->group);
 	return value >= ptr->values[0] && value <= ptr->values[1];
 }
@@ -127,7 +122,7 @@ bool ccs_compare_number_union(const unsigned long value,
 const struct ccs_path_info *ccs_compare_name_union
 (const struct ccs_path_info *name, const struct ccs_name_union *ptr)
 {
-	if (ptr->is_group)
+	if (ptr->group)
 		return ccs_path_matches_group(name, ptr->group);
 	if (ccs_path_matches_pattern(name, ptr->filename))
 		return ptr->filename;
@@ -318,7 +313,7 @@ static bool ccs_check_path2_acl(struct ccs_request_info *r,
 }
 
 /**
- * ccs_check_cmkdev_acl - Check permission for path number number number operation.
+ * ccs_check_mkdev_acl - Check permission for path number number number operation.
  *
  * @r:   Pointer to "struct ccs_request_info".
  * @ptr: Pointer to "struct ccs_acl_info".
@@ -393,7 +388,7 @@ static int ccs_update_path_acl(const u16 perm, struct ccs_acl_param *param)
 		.perm = perm
 	};
 	int error;
-	if (!ccs_parse_name_union(ccs_read_token(param), &e.name))
+	if (!ccs_parse_name_union(param, &e.name))
 		error = -EINVAL;
 	else
 		error = ccs_update_domain(&e.head, sizeof(e), param,
@@ -462,10 +457,10 @@ static int ccs_update_mkdev_acl(const u8 perm, struct ccs_acl_param *param)
 		.perm = perm
 	};
 	int error;
-	if (!ccs_parse_name_union(ccs_read_token(param), &e.name) ||
-	    !ccs_parse_number_union(ccs_read_token(param), &e.mode) ||
-	    !ccs_parse_number_union(ccs_read_token(param), &e.major) ||
-	    !ccs_parse_number_union(ccs_read_token(param), &e.minor))
+	if (!ccs_parse_name_union(param, &e.name) ||
+	    !ccs_parse_number_union(param, &e.mode) ||
+	    !ccs_parse_number_union(param, &e.major) ||
+	    !ccs_parse_number_union(param, &e.minor))
 		error = -EINVAL;
 	else
 		error = ccs_update_domain(&e.head, sizeof(e), param,
@@ -535,8 +530,8 @@ static int ccs_update_path2_acl(const u8 perm, struct ccs_acl_param *param)
 		.perm = perm
 	};
 	int error;
-	if (!ccs_parse_name_union(ccs_read_token(param), &e.name1) ||
-	    !ccs_parse_name_union(ccs_read_token(param), &e.name2))
+	if (!ccs_parse_name_union(param, &e.name1) ||
+	    !ccs_parse_name_union(param, &e.name2))
 		error = -EINVAL;
 	else
 		error = ccs_update_domain(&e.head, sizeof(e), param,
@@ -553,7 +548,7 @@ static int ccs_update_path2_acl(const u8 perm, struct ccs_acl_param *param)
  * @a: Pointer to "struct ccs_acl_info".
  * @b: Pointer to "struct ccs_acl_info".
  *
- * Returns true if @a == @b except permission bits, false otherwise.
+ * Returns true if @a == @b, false otherwise.
  */
 static bool ccs_same_mount_acl(const struct ccs_acl_info *a,
 			       const struct ccs_acl_info *b)
@@ -579,10 +574,10 @@ static int ccs_update_mount_acl(struct ccs_acl_param *param)
 {
 	struct ccs_mount_acl e = { .head.type = CCS_TYPE_MOUNT_ACL };
 	int error;
-	if (!ccs_parse_name_union(ccs_read_token(param), &e.dev_name) ||
-	    !ccs_parse_name_union(ccs_read_token(param), &e.dir_name) ||
-	    !ccs_parse_name_union(ccs_read_token(param), &e.fs_type) ||
-	    !ccs_parse_number_union(ccs_read_token(param), &e.flags))
+	if (!ccs_parse_name_union(param, &e.dev_name) ||
+	    !ccs_parse_name_union(param, &e.dir_name) ||
+	    !ccs_parse_name_union(param, &e.fs_type) ||
+	    !ccs_parse_number_union(param, &e.flags))
 		error = -EINVAL;
 	else
 		error = ccs_update_domain(&e.head, sizeof(e), param,
@@ -758,20 +753,18 @@ static int ccs_new_open_permission(struct file *filp)
 #endif
 
 /**
- * ccs_path_perm - Check permission for "unlink", "rmdir", "truncate", "symlink", "getattr", "chroot" and "unmount".
+ * ccs_path_perm - Check permission for "unlink", "rmdir", "truncate", "symlink", "append", "getattr", "chroot" and "unmount".
  *
  * @operation: Type of operation.
- * @dir:       Pointer to "struct inode". Maybe NULL.
  * @dentry:    Pointer to "struct dentry".
  * @mnt:       Pointer to "struct vfsmount". Maybe NULL.
- * @target:    Symlink's target if @operation is CCS_TYPE_SYMLINK.
+ * @target:    Symlink's target if @operation is CCS_TYPE_SYMLINK,
  *             NULL otherwise.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_perm(const u8 operation, struct inode *dir,
-			 struct dentry *dentry, struct vfsmount *mnt,
-			 const char *target)
+static int ccs_path_perm(const u8 operation, struct dentry *dentry,
+			 struct vfsmount *mnt, const char *target)
 {
 	struct ccs_request_info r;
 	struct ccs_obj_info obj = {
@@ -822,7 +815,6 @@ out:
  * ccs_mkdev_perm - Check permission for "mkblock" and "mkchar".
  *
  * @operation: Type of operation. (CCS_TYPE_MKCHAR or CCS_TYPE_MKBLOCK)
- * @dir:       Pointer to "struct inode".
  * @dentry:    Pointer to "struct dentry".
  * @mnt:       Pointer to "struct vfsmount". Maybe NULL.
  * @mode:      Create mode.
@@ -830,9 +822,9 @@ out:
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_mkdev_perm(const u8 operation, struct inode *dir,
-			  struct dentry *dentry, struct vfsmount *mnt,
-			  const unsigned int mode, unsigned int dev)
+static int ccs_mkdev_perm(const u8 operation, struct dentry *dentry,
+			  struct vfsmount *mnt, const unsigned int mode,
+			  unsigned int dev)
 {
 	struct ccs_request_info r;
 	struct ccs_obj_info obj = {
@@ -880,18 +872,15 @@ out:
  * ccs_path2_perm - Check permission for "rename", "link" and "pivot_root".
  *
  * @operation: Type of operation.
- * @dir1:      Pointer to "struct inode". Maybe NULL.
  * @dentry1:   Pointer to "struct dentry".
  * @mnt1:      Pointer to "struct vfsmount". Maybe NULL.
- * @dir2:      Pointer to "struct inode". Maybe NULL.
  * @dentry2:   Pointer to "struct dentry".
  * @mnt2:      Pointer to "struct vfsmount". Maybe NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path2_perm(const u8 operation, struct inode *dir1,
-			  struct dentry *dentry1, struct vfsmount *mnt1,
-			  struct inode *dir2, struct dentry *dentry2,
+static int ccs_path2_perm(const u8 operation, struct dentry *dentry1,
+			  struct vfsmount *mnt1, struct dentry *dentry2,
 			  struct vfsmount *mnt2)
 {
 	struct ccs_request_info r;
@@ -1007,8 +996,8 @@ static int ccs_update_path_number_acl(const u8 perm,
 		.perm = perm
 	};
 	int error;
-	if (!ccs_parse_name_union(ccs_read_token(param), &e.name) ||
-	    !ccs_parse_number_union(ccs_read_token(param), &e.number))
+	if (!ccs_parse_name_union(param, &e.name) ||
+	    !ccs_parse_number_union(param, &e.number))
 		error = -EINVAL;
 	else
 		error = ccs_update_domain(&e.head, sizeof(e), param,
@@ -1023,16 +1012,14 @@ static int ccs_update_path_number_acl(const u8 perm,
  * ccs_path_number_perm - Check permission for "create", "mkdir", "mkfifo", "mksock", "ioctl", "chmod", "chown", "chgrp".
  *
  * @type:   Type of operation.
- * @dir:    Pointer to "struct inode". Maybe NULL.
  * @dentry: Pointer to "struct dentry".
  * @vfsmnt: Pointer to "struct vfsmount". Maybe NULL.
  * @number: Number.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_number_perm(const u8 type, struct inode *dir,
-				struct dentry *dentry, struct vfsmount *vfsmnt,
-				unsigned long number)
+static int ccs_path_number_perm(const u8 type, struct dentry *dentry,
+				struct vfsmount *vfsmnt, unsigned long number)
 {
 	struct ccs_request_info r;
 	struct ccs_obj_info obj = {
@@ -1081,7 +1068,7 @@ out:
 static int __ccs_ioctl_permission(struct file *filp, unsigned int cmd,
 				  unsigned long arg)
 {
-	return ccs_path_number_perm(CCS_TYPE_IOCTL, NULL, filp->f_dentry,
+	return ccs_path_number_perm(CCS_TYPE_IOCTL, filp->f_dentry,
 				    filp->f_vfsmnt, cmd);
 }
 
@@ -1099,7 +1086,7 @@ static int __ccs_chmod_permission(struct dentry *dentry,
 {
 	if (mode == (mode_t) -1)
 		return 0;
-	return ccs_path_number_perm(CCS_TYPE_CHMOD, NULL, dentry, vfsmnt,
+	return ccs_path_number_perm(CCS_TYPE_CHMOD, dentry, vfsmnt,
 				    mode & S_IALLUGO);
 }
 
@@ -1121,11 +1108,11 @@ static int __ccs_chown_permission(struct dentry *dentry,
 	if (user == (uid_t) -1 && group == (gid_t) -1)
 		return 0;
 	if (user != (uid_t) -1)
-		error = ccs_path_number_perm(CCS_TYPE_CHOWN, NULL, dentry,
-					     vfsmnt, user);
+		error = ccs_path_number_perm(CCS_TYPE_CHOWN, dentry, vfsmnt,
+					     user);
 	if (!error && group != (gid_t) -1)
-		error = ccs_path_number_perm(CCS_TYPE_CHGRP, NULL, dentry,
-					     vfsmnt, group);
+		error = ccs_path_number_perm(CCS_TYPE_CHGRP, dentry, vfsmnt,
+					     group);
 	return error;
 }
 
@@ -1144,11 +1131,14 @@ static int __ccs_fcntl_permission(struct file *file, unsigned int cmd,
 	if (!(cmd == F_SETFL && ((arg ^ file->f_flags) & O_APPEND)))
 		return 0;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 33)
-	/* 01 means "write". */
-	return __ccs_open_permission(file->f_dentry, file->f_vfsmnt, 01);
+	return __ccs_open_permission(file->f_dentry, file->f_vfsmnt,
+				     O_WRONLY | (arg & O_APPEND));
+#elif defined(RHEL_MAJOR) && RHEL_MAJOR == 6
+	return __ccs_open_permission(file->f_dentry, file->f_vfsmnt,
+				     O_WRONLY | (arg & O_APPEND));
 #else
-	/* 02 means "write". */
-	return __ccs_open_permission(file->f_dentry, file->f_vfsmnt, 02);
+	return __ccs_open_permission(file->f_dentry, file->f_vfsmnt,
+				     (O_WRONLY + 1) | (arg & O_APPEND));
 #endif
 }
 
@@ -1163,9 +1153,8 @@ static int __ccs_fcntl_permission(struct file *file, unsigned int cmd,
 static int __ccs_pivot_root_permission(struct path *old_path,
 				       struct path *new_path)
 {
-	return ccs_path2_perm(CCS_TYPE_PIVOT_ROOT, NULL, new_path->dentry,
-			      new_path->mnt, NULL, old_path->dentry,
-			      old_path->mnt);
+	return ccs_path2_perm(CCS_TYPE_PIVOT_ROOT, new_path->dentry,
+			      new_path->mnt, old_path->dentry, old_path->mnt);
 }
 
 /**
@@ -1177,8 +1166,7 @@ static int __ccs_pivot_root_permission(struct path *old_path,
  */
 static int __ccs_chroot_permission(struct path *path)
 {
-	return ccs_path_perm(CCS_TYPE_CHROOT, NULL, path->dentry, path->mnt,
-			     NULL);
+	return ccs_path_perm(CCS_TYPE_CHROOT, path->dentry, path->mnt, NULL);
 }
 
 /**
@@ -1191,7 +1179,7 @@ static int __ccs_chroot_permission(struct path *path)
  */
 static int __ccs_umount_permission(struct vfsmount *mnt, int flags)
 {
-	return ccs_path_perm(CCS_TYPE_UMOUNT, NULL, mnt->mnt_root, mnt, NULL);
+	return ccs_path_perm(CCS_TYPE_UMOUNT, mnt->mnt_root, mnt, NULL);
 }
 
 /**
@@ -1200,6 +1188,8 @@ static int __ccs_umount_permission(struct vfsmount *mnt, int flags)
  * @param: Pointer to "struct ccs_acl_param".
  *
  * Returns 0 on success, negative value otherwise.
+ *
+ * Caller holds ccs_read_lock().
  */
 int ccs_write_file(struct ccs_acl_param *param)
 {
@@ -1235,7 +1225,6 @@ int ccs_write_file(struct ccs_acl_param *param)
 /**
  * __ccs_mknod_permission - Check permission for vfs_mknod().
  *
- * @dir:    Pointer to "struct inode".
  * @dentry: Pointer to "struct dentry".
  * @mnt:    Pointer to "struct vfsmount". Maybe NULL.
  * @mode:   Device type and permission.
@@ -1243,32 +1232,31 @@ int ccs_write_file(struct ccs_acl_param *param)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_mknod_permission(struct inode *dir, struct dentry *dentry,
-				  struct vfsmount *mnt,
+static int __ccs_mknod_permission(struct dentry *dentry, struct vfsmount *mnt,
 				  const unsigned int mode, unsigned int dev)
 {
 	int error = 0;
 	const unsigned int perm = mode & S_IALLUGO;
 	switch (mode & S_IFMT) {
 	case S_IFCHR:
-		error = ccs_mkdev_perm(CCS_TYPE_MKCHAR, dir, dentry, mnt, perm,
+		error = ccs_mkdev_perm(CCS_TYPE_MKCHAR, dentry, mnt, perm,
 				       dev);
 		break;
 	case S_IFBLK:
-		error = ccs_mkdev_perm(CCS_TYPE_MKBLOCK, dir, dentry, mnt,
-				       perm, dev);
+		error = ccs_mkdev_perm(CCS_TYPE_MKBLOCK, dentry, mnt, perm,
+				       dev);
 		break;
 	case S_IFIFO:
-		error = ccs_path_number_perm(CCS_TYPE_MKFIFO, dir, dentry, mnt,
+		error = ccs_path_number_perm(CCS_TYPE_MKFIFO, dentry, mnt,
 					     perm);
 		break;
 	case S_IFSOCK:
-		error = ccs_path_number_perm(CCS_TYPE_MKSOCK, dir, dentry, mnt,
+		error = ccs_path_number_perm(CCS_TYPE_MKSOCK, dentry, mnt,
 					     perm);
 		break;
 	case 0:
 	case S_IFREG:
-		error = ccs_path_number_perm(CCS_TYPE_CREATE, dir, dentry, mnt,
+		error = ccs_path_number_perm(CCS_TYPE_CREATE, dentry, mnt,
 					     perm);
 		break;
 	}
@@ -1278,47 +1266,42 @@ static int __ccs_mknod_permission(struct inode *dir, struct dentry *dentry,
 /**
  * __ccs_mkdir_permission - Check permission for vfs_mkdir().
  *
- * @dir:    Pointer to "struct inode".
  * @dentry: Pointer to "struct dentry".
  * @mnt:    Pointer to "struct vfsmount". Maybe NULL.
  * @mode:   Create mode.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_mkdir_permission(struct inode *dir, struct dentry *dentry,
-				  struct vfsmount *mnt, unsigned int mode)
+static int __ccs_mkdir_permission(struct dentry *dentry, struct vfsmount *mnt,
+				  unsigned int mode)
 {
-	return ccs_path_number_perm(CCS_TYPE_MKDIR, dir, dentry, mnt, mode);
+	return ccs_path_number_perm(CCS_TYPE_MKDIR, dentry, mnt, mode);
 }
 
 /**
  * __ccs_rmdir_permission - Check permission for vfs_rmdir().
  *
- * @dir:    Pointer to "struct inode".
  * @dentry: Pointer to "struct dentry".
  * @mnt:    Pointer to "struct vfsmount". Maybe NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_rmdir_permission(struct inode *dir, struct dentry *dentry,
-				  struct vfsmount *mnt)
+static int __ccs_rmdir_permission(struct dentry *dentry, struct vfsmount *mnt)
 {
-	return ccs_path_perm(CCS_TYPE_RMDIR, dir, dentry, mnt, NULL);
+	return ccs_path_perm(CCS_TYPE_RMDIR, dentry, mnt, NULL);
 }
 
 /**
  * __ccs_unlink_permission - Check permission for vfs_unlink().
  *
- * @dir:    Pointer to "struct inode".
  * @dentry: Pointer to "struct dentry".
  * @mnt:    Pointer to "struct vfsmount". Maybe NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_unlink_permission(struct inode *dir, struct dentry *dentry,
-				   struct vfsmount *mnt)
+static int __ccs_unlink_permission(struct dentry *dentry, struct vfsmount *mnt)
 {
-	return ccs_path_perm(CCS_TYPE_UNLINK, dir, dentry, mnt, NULL);
+	return ccs_path_perm(CCS_TYPE_UNLINK, dentry, mnt, NULL);
 }
 
 /**
@@ -1332,23 +1315,22 @@ static int __ccs_unlink_permission(struct inode *dir, struct dentry *dentry,
 static int __ccs_getattr_permission(struct vfsmount *mnt,
 				    struct dentry *dentry)
 {
-	return ccs_path_perm(CCS_TYPE_GETATTR, NULL, dentry, mnt, NULL);
+	return ccs_path_perm(CCS_TYPE_GETATTR, dentry, mnt, NULL);
 }
 
 /**
  * __ccs_symlink_permission - Check permission for vfs_symlink().
  *
- * @dir:    Pointer to "struct inode".
  * @dentry: Pointer to "struct dentry".
  * @mnt:    Pointer to "struct vfsmount". Maybe NULL.
  * @from:   Content of symlink.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_symlink_permission(struct inode *dir, struct dentry *dentry,
+static int __ccs_symlink_permission(struct dentry *dentry,
 				    struct vfsmount *mnt, const char *from)
 {
-	return ccs_path_perm(CCS_TYPE_SYMLINK, dir, dentry, mnt, from);
+	return ccs_path_perm(CCS_TYPE_SYMLINK, dentry, mnt, from);
 }
 
 /**
@@ -1362,47 +1344,40 @@ static int __ccs_symlink_permission(struct inode *dir, struct dentry *dentry,
 static int __ccs_truncate_permission(struct dentry *dentry,
 				     struct vfsmount *mnt)
 {
-	return ccs_path_perm(CCS_TYPE_TRUNCATE, NULL, dentry, mnt, NULL);
+	return ccs_path_perm(CCS_TYPE_TRUNCATE, dentry, mnt, NULL);
 }
 
 /**
  * __ccs_rename_permission - Check permission for vfs_rename().
  *
- * @old_dir:    Pointer to "struct inode".
  * @old_dentry: Pointer to "struct dentry".
- * @new_dir:    Pointer to "struct inode".
  * @new_dentry: Pointer to "struct dentry".
  * @mnt:        Pointer to "struct vfsmount". Maybe NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int __ccs_rename_permission(struct inode *old_dir,
-				   struct dentry *old_dentry,
-				   struct inode *new_dir,
+static int __ccs_rename_permission(struct dentry *old_dentry,
 				   struct dentry *new_dentry,
 				   struct vfsmount *mnt)
 {
-	return ccs_path2_perm(CCS_TYPE_RENAME, old_dir, old_dentry, mnt,
-			      new_dir, new_dentry, mnt);
+	return ccs_path2_perm(CCS_TYPE_RENAME, old_dentry, mnt, new_dentry,
+			      mnt);
 }
 
 /**
  * __ccs_link_permission - Check permission for vfs_link().
  *
  * @old_dentry: Pointer to "struct dentry".
- * @new_dir:    Pointer to "struct inode".
  * @new_dentry: Pointer to "struct dentry".
  * @mnt:        Pointer to "struct vfsmount". Maybe NULL.
  *
  * Returns 0 on success, negative value otherwise.
  */
 static int __ccs_link_permission(struct dentry *old_dentry,
-				 struct inode *new_dir,
 				 struct dentry *new_dentry,
 				 struct vfsmount *mnt)
 {
-	return ccs_path2_perm(CCS_TYPE_LINK, NULL, old_dentry, mnt,
-			      new_dir, new_dentry, mnt);
+	return ccs_path2_perm(CCS_TYPE_LINK, old_dentry, mnt, new_dentry, mnt);
 }
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 30)
@@ -1419,8 +1394,7 @@ static int __ccs_open_exec_permission(struct dentry *dentry,
 				      struct vfsmount *mnt)
 {
 	return (ccs_current_flags() & CCS_TASK_IS_IN_EXECVE) ?
-		/* 01 means "read". */
-		__ccs_open_permission(dentry, mnt, 01) : 0;
+		__ccs_open_permission(dentry, mnt, O_RDONLY + 1) : 0;
 }
 
 /**
@@ -1433,8 +1407,7 @@ static int __ccs_open_exec_permission(struct dentry *dentry,
  */
 static int __ccs_uselib_permission(struct dentry *dentry, struct vfsmount *mnt)
 {
-	/* 01 means "read". */
-	return __ccs_open_permission(dentry, mnt, 01);
+	return __ccs_open_permission(dentry, mnt, O_RDONLY + 1);
 }
 
 #endif
